# Copyright 2019 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus_ext_newrelic.trace import NewRelicTraceExporter
from opencensus_ext_newrelic.stats import NewRelicStatsExporter

try:
    from opencensus_ext_newrelic.version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"  # pragma: no cover

__all__ = ("NewRelicTraceExporter", "NewRelicStatsExporter")



import os
import functools
import pytest
from requests.adapters import HTTPAdapter
from requests.models import Response

try:
    string_types = basestring
except NameError:
    string_types = str


def _ensure_utf8(s):
    if not isinstance(s, string_types):
        try:
            s = s.decode("utf-8")
        except Exception:
            return
    return s


def _mocked_status_code(wrapped, status_code, disable_requests):
    if disable_requests:

        @functools.wraps(wrapped)
        def wrapper(self, request, *args, **kwargs):
            response = Response()
            response.status_code = status_code
            response.request = request
            return response

    else:

        @functools.wraps(wrapped)
        def wrapper(*args, **kwargs):
            response = wrapped(*args, **kwargs)
            response.status_code = status_code
            return response

    return wrapper


def _http_response(status_code):
    return status_code


@pytest.fixture
def ensure_utf8():
    return _ensure_utf8


@pytest.fixture
def insert_key():
    return os.environ.get("NEW_RELIC_INSERT_KEY", "")


@pytest.fixture
def hosts():
    host = os.environ.get("NEW_RELIC_HOST", "")

    hosts = {"trace": None, "metric": None}

    if host.startswith("staging"):
        hosts["trace"] = "staging-trace-api.newrelic.com"
        hosts["metric"] = "staging-metric-api.newrelic.com"

    return hosts


@pytest.fixture(autouse=True)
def mocked_responses(request, monkeypatch):
    status_code = None
    disable_requests = False

    # Determine if the proper key exists for this test. If it does not exist,
    # do not attempt to send requests to New Relic and instead immediately
    # return a response object with the mocked status code.
    if "insert_key" in request.node.fixturenames:
        insert_key = request.getfixturevalue("insert_key")
        if not insert_key:
            status_code = 200
            disable_requests = True

    # The response marker always overrides the status code
    http_response_marker = request.node.get_closest_marker("http_response")
    if http_response_marker:
        status_code = _http_response(
            *http_response_marker.args, **http_response_marker.kwargs
        )

    # If we should not inject a mocked response, return without patching
    if status_code is None:
        return

    wrapped = HTTPAdapter.send
    monkeypatch.setattr(
        HTTPAdapter, "send", _mocked_status_code(wrapped, status_code, disable_requests)
    )



import logging
import json
import pytest
from datetime import datetime
from opencensus.stats import aggregation as aggregation_module
from opencensus.stats import measure as measure_module
from opencensus.tags import tag_map as tag_map_module
from opencensus.stats import view as view_module
from opencensus.stats import view_data as view_data_module
from opencensus.stats import metric_utils
from opencensus_ext_newrelic import NewRelicStatsExporter


# The latency in milliseconds
MEASURE = measure_module.MeasureFloat("number", "A number!", "things")

GAUGE_VIEWS = {
    "last": view_module.View(
        "last",
        "A last value",
        ("tag",),
        MEASURE,
        aggregation_module.LastValueAggregation(),
    )
}
COUNT_VIEWS = {
    "count": view_module.View(
        "count", "A count", ("tag",), MEASURE, aggregation_module.CountAggregation()
    ),
    "sum": view_module.View(
        "sum", "A sum", ("tag",), MEASURE, aggregation_module.SumAggregation()
    ),
}
DISTRIBUTION_VIEWS = {
    "distribution": view_module.View(
        "distribution",
        "A distribution",
        ("tag",),
        MEASURE,
        aggregation_module.DistributionAggregation([50.0, 200.0]),
    )
}
VIEWS = {}
VIEWS.update(GAUGE_VIEWS)
VIEWS.update(COUNT_VIEWS)
VIEWS.update(DISTRIBUTION_VIEWS)

TEST_TIMESTAMP = datetime.utcfromtimestamp(1557785724.128)


@pytest.fixture
def stats_exporter(insert_key, hosts):
    exporter = NewRelicStatsExporter(
        insert_key, host=hosts["metric"], service_name="Python Application"
    )
    exporter._thread.cancel()

    for view in VIEWS.values():
        exporter.on_register_view(view)

    return exporter


def record_values(view_data_objects, tags, value=1, count=1):
    tag_map = tag_map_module.TagMap(tags)
    for view_data_object in view_data_objects:
        for _ in range(count):
            # Timestamp here is only used to record exemplars. It is safe to
            # leave it as None
            view_data_object.record(tag_map, value, None)


def generate_metrics(view_data_objects):
    metrics = []
    for view_data_object in view_data_objects:
        metric = metric_utils.view_data_to_metric(view_data_object, TEST_TIMESTAMP)
        metrics.append(metric)
    return metrics


def to_view_data(view):
    # Start and end time should be the same value
    # https://github.com/census-instrumentation/opencensus-python/blob/b28a83f84dbbfb539c90c8844a96e9394df24c5b/opencensus/stats/measure_to_view_map.py#L105L106
    #
    # The start and end times should not be used in the calculation of metric
    # timing information
    view_data = view_data_module.ViewData(
        view=view,
        start_time="2019-05-11T00:07:45.0Z",
        end_time="2019-05-11T00:07:45.0Z",
    )
    return view_data


@pytest.mark.parametrize(
    "tag_values",
    ((None,), ("foo",), ("foo", "bar")),
    ids=("no tag value", "single tag value", "multiple tag values"),
)
def test_stats(stats_exporter, ensure_utf8, tag_values):
    view_data_objects = [to_view_data(view) for view in VIEWS.values()]

    # Record values
    for tag_value in tag_values:
        if tag_value:
            tags = {"tag": tag_value}
        else:
            tags = None

        record_values(view_data_objects, tags, value=100)

    # Generate metrics
    metrics = generate_metrics(view_data_objects)

    # Send metrics to exporter
    response = stats_exporter.export_metrics(metrics)
    data = json.loads(ensure_utf8(response.request.body))

    common = data[0]["common"]
    assert len(common) == 2
    assert common["interval.ms"] == stats_exporter.interval * 1000
    assert common["attributes"]["service.name"] == "Python Application"

    metrics_data = data[0]["metrics"]
    assert len(metrics_data) == (
        len(tag_values) * (len(GAUGE_VIEWS) + len(COUNT_VIEWS))
    ), metrics_data

    remaining_tags = {
        view: set(tag_values) for view in VIEWS if view not in DISTRIBUTION_VIEWS
    }

    for metric_data in metrics_data:
        view_name = metric_data["name"]
        view = VIEWS[view_name]

        # Distribution views will raise an exception since there will be no
        # tags expected for a distribution aggregation
        expected_tags = remaining_tags[view.name]

        assert metric_data["attributes"]["measure.name"] == MEASURE.name
        assert metric_data["attributes"]["measure.unit"] == MEASURE.unit
        assert metric_data["timestamp"] == 1557785724128

        if view_name in GAUGE_VIEWS:
            # Check that each metric is a gauge
            assert "type" not in metric_data
        else:
            assert metric_data["type"] == "count"

        expected_tags.remove(metric_data["attributes"]["tag"])
        if not expected_tags:
            remaining_tags.pop(view.name)

    assert not remaining_tags


def test_empty_payload_is_not_sent(stats_exporter):
    response = stats_exporter.export_metrics(())
    assert response is None


def test_send_metrics_exception(stats_exporter, caplog):
    # Remove the client object to force an exception when send_metrics is called
    delattr(stats_exporter, "client")
    view_data = to_view_data(VIEWS["last"])
    view_data.record(None, 100, None)

    metric = metric_utils.view_data_to_metric(view_data, TEST_TIMESTAMP)

    response = stats_exporter.export_metrics([metric])
    assert response is None

    assert (
        "opencensus_ext_newrelic.stats",
        logging.ERROR,
        "New Relic send_metrics failed with an exception.",
    ) in caplog.record_tuples


@pytest.mark.http_response(status_code=500)
def test_bad_http_response(stats_exporter, caplog):
    view_data = to_view_data(VIEWS["last"])
    view_data.record(None, 100, None)

    metric = metric_utils.view_data_to_metric(view_data, TEST_TIMESTAMP)

    response = stats_exporter.export_metrics([metric])
    assert response.status_code == 500

    assert (
        "opencensus_ext_newrelic.stats",
        logging.ERROR,
        "New Relic send_metrics failed with status code: 500",
    ) in caplog.record_tuples


def test_metric_computes_delta(stats_exporter, ensure_utf8):
    view_data_objects = [to_view_data(view) for view in COUNT_VIEWS.values()]
    for delta_first, delta_second in ((2, 1), (1, 0), (0, 0)):
        record_values(view_data_objects, {"tag": "first"}, count=delta_first)
        record_values(view_data_objects, {"tag": "second"}, count=delta_second)
        metrics = generate_metrics(view_data_objects)

        response = stats_exporter.export_metrics(metrics)
        data = json.loads(ensure_utf8(response.request.body))

        expected_values = {
            view_name: {"first": delta_first, "second": delta_second}
            for view_name in COUNT_VIEWS
        }
        metrics_data = data[0]["metrics"]
        assert len(metrics_data) == 4
        for metric in metrics_data:
            assert metric["type"] == "count"
            view_values = expected_values[metric["name"]]
            expected_value = view_values.pop(metric["attributes"]["tag"])
            assert metric["value"] == expected_value
            if not view_values:
                expected_values.pop(metric["name"])
        assert not expected_values


def test_stop_clears_all_state(stats_exporter):
    stats_exporter.stop()

    assert stats_exporter.client is None
    assert stats_exporter.views is None

    # Call on_register_view after stop
    for view in VIEWS.values():
        stats_exporter.on_register_view(view)


def test_default_exporter_values(insert_key):
    exporter = NewRelicStatsExporter(insert_key, service_name="Python Application")
    exporter._thread.cancel()

    assert exporter.interval == 5



# Copyright 2019 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opencensus.common.transports import async_
from opencensus.common.utils import timestamp_to_microseconds
from opencensus.trace import base_exporter
from newrelic_telemetry_sdk import Span, SpanClient

import logging

_logger = logging.getLogger(__name__)


class DefaultTransport(async_.AsyncTransport):
    def __init__(
        self, exporter, grace_period=None, max_batch_size=600, wait_period=5.0
    ):
        super(DefaultTransport, self).__init__(
            exporter, grace_period, max_batch_size, wait_period
        )

    def stop(self):
        """Terminate the background thread"""
        self.worker._export_pending_data()


class NewRelicTraceExporter(base_exporter.Exporter):
    """Export Span data to the New Relic platform

    This class is responsible for marshalling trace data to the New Relic
    platform.

    :param insert_key: Insights insert key
    :type insert_key: str
    :param service_name: (optional) The name of the entity to report spans
        into. Defaults to "Python Application".
    :type service_name: str
    :param host: (optional) Override the host for the API endpoint.
    :type host: str
    :param transport: (optional) Class for creating new transport objects. It
        should extend from the base_exporter
        :class:`opencensus.common.transports.base.Transport` type and implement
        :meth:`opencensus.common.transports.base.Transport.export`. Defaults to
        an async transport sending data every 5 seconds. The other option is
        :class:`opencensus.common.transports.async.AsyncTransport`.
    :type transport: :class:`opencensus.common.transports.base.Transport`

    Usage::

        >>> import os
        >>> from opencensus_ext_newrelic import NewRelicTraceExporter
        >>> insert_key = os.environ.get("NEW_RELIC_INSERT_KEY")
        >>> trace_exporter = NewRelicTraceExporter(
        ...     insert_key, service_name="My Service")
        >>> trace_exporter.stop()
    """

    def __init__(self, insert_key, service_name, host=None, transport=DefaultTransport):
        self._common = {"attributes": {"service.name": service_name}}
        self.client = SpanClient(insert_key=insert_key, host=host)
        self._transport = transport(self)

    def emit(self, span_datas):
        """Immediately marshal span data to the tracing backend

        :param span_datas: list of :class:`opencensus.trace.span_data.SpanData`
            to emit
        :type span_datas: list
        """
        spans = []
        for span_data in span_datas:
            start_timestamp_mus = timestamp_to_microseconds(span_data.start_time)
            end_timestamp_mus = timestamp_to_microseconds(span_data.end_time)
            duration_mus = end_timestamp_mus - start_timestamp_mus

            start_time_ms = start_timestamp_mus // 1000
            duration_ms = duration_mus // 1000

            span = Span(
                name=span_data.name,
                tags=span_data.attributes,
                guid=span_data.span_id,
                trace_id=span_data.context.trace_id,
                parent_id=span_data.parent_span_id,
                start_time_ms=start_time_ms,
                duration_ms=duration_ms,
            )

            spans.append(span)

        try:
            response = self.client.send_batch(spans, self._common)
        except Exception:
            _logger.exception("New Relic send_spans failed with an exception.")
            return

        if not response.ok:
            _logger.error(
                "New Relic send_spans failed with status code: %r", response.status_code
            )

        return response

    def export(self, span_datas):
        """Export the trace. Send trace to transport, and transport will call
        exporter.emit() to actually send the trace to the specified tracing
        backend.

        :param span_datas: list of :class:`opencensus.trace.span_data.SpanData`
            to export.
        :type span_datas: list
        """
        if self._transport is not None:
            return self._transport.export(span_datas)

    def stop(self):
        """Terminate the exporter and any background threads"""
        transport = self._transport

        # Send all pending data
        if hasattr(transport, "stop"):
            transport.stop()

        # Clear all internal state
        self._transport = self.client = None



import logging
import json
import pytest
from opencensus_ext_newrelic import NewRelicTraceExporter
from opencensus.common.transports import sync
from opencensus.trace import span_context
from opencensus.trace.span_data import SpanData


class Transport(sync.SyncTransport):
    def export(self, datas):
        return self.exporter.emit(datas)


@pytest.fixture
def trace_exporter(hosts, insert_key):
    exporter = NewRelicTraceExporter(
        insert_key=insert_key,
        transport=Transport,
        host=hosts["trace"],
        service_name="Python Application",
    )
    return exporter


SPAN_DATA = {
    "name": "test_span",
    "context": span_context.SpanContext(
        trace_id="2dd43a1d6b2549c6bc2a1a54c2fc0b05", span_id="6e0c63257de34c92"
    ),
    "span_id": "6e0c63257de34c92",
    "parent_span_id": "6e0c63257de34c93",
    "attributes": {"key1": "value1"},
    "start_time": "2019-05-11T00:07:48.0Z",
    "end_time": "2019-05-11T00:07:49.0Z",
    "span_kind": 0,
}
for field in SpanData._fields:
    if field not in SPAN_DATA:
        SPAN_DATA[field] = None

SPAN_DATA = SpanData(**SPAN_DATA)


def test_trace(trace_exporter, ensure_utf8):
    duration = 1000
    timestamp = 1557533268000

    response = trace_exporter.export([SPAN_DATA])
    data = json.loads(ensure_utf8(response.request.body))
    assert len(data) == 1
    data = data[0]
    spans, common = data["spans"], data["common"]
    assert len(spans) == 1
    span = spans[0]
    attributes = span["attributes"]

    assert span["id"] == SPAN_DATA.context.span_id
    assert span["timestamp"] == timestamp
    assert span["trace.id"] == SPAN_DATA.context.trace_id

    for name, value in SPAN_DATA.attributes.items():
        assert attributes[name] == value

    assert attributes["duration.ms"] == duration
    assert attributes["name"] == SPAN_DATA.name
    assert attributes["parent.id"] == SPAN_DATA.parent_span_id

    assert common["attributes"]["service.name"] == "Python Application"


def test_send_spans_exception(trace_exporter, caplog):
    # Remove the client object to force an exception when send_spans is called
    delattr(trace_exporter, "client")

    response = trace_exporter.export([SPAN_DATA])
    assert response is None

    assert (
        "opencensus_ext_newrelic.trace",
        logging.ERROR,
        "New Relic send_spans failed with an exception.",
    ) in caplog.record_tuples


@pytest.mark.http_response(status_code=500)
def test_bad_http_response(trace_exporter, caplog):
    trace_exporter.export([SPAN_DATA])

    assert (
        "opencensus_ext_newrelic.trace",
        logging.ERROR,
        "New Relic send_spans failed with status code: 500",
    ) in caplog.record_tuples


def test_stop_clears_all_state(trace_exporter):
    trace_exporter.stop()

    assert trace_exporter.client is None

    # Calling export after stop
    trace_exporter.export(())



import setuptools

setuptools.setup(
    use_scm_version={
        "write_to": "src/opencensus_ext_newrelic/version.py",
        "write_to_template": """\
# coding: utf-8
# file generated by setuptools_scm
# don't change, don't track in version control
version = "{version!s}"
""",
    },
    install_requires=("opencensus~=0.7", "newrelic-telemetry-sdk~=0.1"),
)



# Copyright 2019 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
from opencensus.stats import stats
from opencensus.metrics import transport
from opencensus.stats import aggregation
from newrelic_telemetry_sdk import MetricClient, GaugeMetric, CountMetric

import logging

_logger = logging.getLogger(__name__)
COUNT_AGGREGATION_TYPES = {aggregation.CountAggregation, aggregation.SumAggregation}


def create_identity(name, tags):
    tags = frozenset(tags.items())
    identity = (name, tags)
    return identity


class NewRelicStatsExporter(object):
    """Export Metric data to the New Relic platform

    This class is responsible for marshalling metric data to the New Relic
    platform.

    :param insert_key: Insights insert key
    :type insert_key: str
    :param interval: (optional) Metrics will be sent every ``interval``
        seconds. Default is 5 seconds.
    :type interval: int or float
    :param host: (optional) Override the host for the API endpoint.
    :type host: str

    Usage::

        >>> import os
        >>> from opencensus_ext_newrelic import NewRelicStatsExporter
        >>> insert_key = os.environ.get("NEW_RELIC_INSERT_KEY")
        >>> stats_exporter = NewRelicStatsExporter(
        ...     insert_key, service_name="My Service")
        >>> stats_exporter.stop()
    """

    def __init__(self, insert_key, service_name, host=None, interval=5):
        self.client = MetricClient(insert_key=insert_key, host=host)
        self.views = {}
        self.count_values = {}

        # Register an exporter thread for this exporter
        thread = self._thread = transport.get_exporter_thread(
            [stats.stats], self, interval=interval
        )
        self.interval = thread.interval

        self._common = {
            "interval.ms": self.interval * 1000,
            "attributes": {"service.name": service_name},
        }

    def on_register_view(self, view):
        """Called when a view is registered with the view manager

        :param view: View object to register
        :type view: :class:`opencensus.stats.view.View`
        """
        if self.views is not None:
            self.views[view.name] = view

    def export_metrics(self, metrics):
        """Immediately send all metric data to the monitoring backend.

        :param metrics: list of Metric objects to send to the monitoring
            backend
        :type metrics: :class:`opencensus.metrics.export.metric.Metric`
        """
        nr_metrics = []
        for metric in metrics:
            descriptor = metric.descriptor
            name = descriptor.name
            view = self.views[name]
            measure_name = view.measure.name
            measure_unit = view.measure.unit
            aggregation_type = view.aggregation

            tags = {"measure.name": measure_name, "measure.unit": measure_unit}

            for timeseries in metric.time_series:
                # In distribution aggregations, the values do not have a value attribute
                # We simply ignore this case for now
                try:
                    value = timeseries.points[0].value.value
                except AttributeError:
                    break

                timestamp = timeseries.points[0].timestamp
                time_tuple = timestamp.utctimetuple()
                epoch_time_secs = calendar.timegm(time_tuple)
                epoch_time_mus = epoch_time_secs * 1e6 + timestamp.microsecond
                end_time_ms = epoch_time_mus // 1000

                labels = (
                    (k, l.value) for k, l in zip(view.columns, timeseries.label_values)
                )

                _tags = tags.copy()
                _tags.update(labels)

                if type(aggregation_type) in COUNT_AGGREGATION_TYPES:
                    identity = create_identity(name, _tags)

                    # Compute a delta count based on the previous value. If one
                    # does not exist, report the raw count value.
                    delta = value - self.count_values.get(identity, 0)
                    self.count_values[identity] = value
                    value = delta

                    nr_metric = CountMetric(
                        name=name, value=value, tags=_tags, end_time_ms=end_time_ms
                    )

                else:
                    nr_metric = GaugeMetric(
                        name=name, value=value, tags=_tags, end_time_ms=end_time_ms
                    )

                nr_metrics.append(nr_metric)

        # Do not send an empty metrics payload (only DistributionAggregation values)
        if not nr_metrics:
            return

        try:
            response = self.client.send_batch(nr_metrics, common=self._common)
        except Exception:
            _logger.exception("New Relic send_metrics failed with an exception.")
            return

        if not response.ok:
            _logger.error(
                "New Relic send_metrics failed with status code: %r",
                response.status_code,
            )
        return response

    def stop(self):
        """Terminate the exporter background thread"""
        thread = self._thread

        stop = getattr(thread, "stop", None) or getattr(thread, "cancel")
        stop()

        # Send all pending metrics
        thread.function(*thread.args, **thread.kwargs)

        # Clear all internal state
        self._thread = self.client = self.views = self.count_values = None
